"""Module providing slideshow video class."""

from kivy.graphics import PushMatrix, PopMatrix, Rotate, Color, Rectangle
from kivy.logger import Logger
from kivy.uix.video import Video
from kivy.uix.widget import Widget

from .base import LabeledContent


class SlideshowVideo(LabeledContent):
    """Video slideshow widget.

    Loads the video from the specified File and starts playing it as soon as the
    widget becomes visible. The video is scaled to fit the entire widget,
    respecting the aspect ratio.
    """

    def __init__(self, file, config):
        """Initialize slideshow video instance.

        :param file: Repository file instance for the video to be displayed.
        :type file: repository.File
        :param config: Dictionary with the following entries:
            rotation: Angle in degrees (int) by which the video is rotated clockwise.
            bgolor: Canvas background color (list(3)) for areas, which are not covered by the video.
            resize: Mode (str) for resizing of videos. Must equal "fit" or "fill".
        :type config: dict
        """
        super().__init__(file, config)
        # Calculate total rotation from image rotation and rotation configured for the slideshow.
        # To avoid negative values and values beyond 360°, we additionally apply a modulo operation.
        self._rotation = (file.rotation - config['rotation']) % 360
        self._bgcolor = config['bg_color']
        self._resize = config['resize']
        # Create and add video widget.
        self._video = Video(source=file.source, state='stop', fit_mode='contain', options={'eos': 'loop'})
        if not config['sound']: self._video.volume = 0
        self.add_widget(self._video, len(self.children))
        # Call update_canvas method when the size of the widget changes.
        self.bind(size=self.update_canvas)
        # Call autoplay method when the widget becomes visible/invisible.
        self.bind(parent=self.autoplay)

    def autoplay(self, *args):
        """Start/stop playing the video when the widget becomes visible/invisible."""
        if self.parent is None:
            self.stop()
        else:
            self.play()

    def play(self):
        """Start playing content."""
        self._video.state = 'play'

    def stop(self):
        """Stop playing content."""
        self._video.state = 'stop'

    def update_canvas(self, *args):
        """Update canvas when the size of the widget changes."""
        # Clear before and after groups of video canvas.
        self._video.canvas.before.clear()
        self._video.canvas.after.clear()

        # Fill canvas with background color.
        with self._video.canvas.before:
            Color(*self._bgcolor)
            Rectangle(pos=(0, 0), size=self.size)

        # Determine widget and video aspect ratios.
        widget_ratio = self.width/self.height
        # We need to rely on the file meta data in this case since the Kivy
        # video class does not have a video_ratio attribute and the
        # dimensions of the video widget have not been adjusted yet.
        if self._file.width > 0 and self._file.height > 0:
            video_ratio = self._file.width/self._file.height
        else:
            video_ratio = 16/9
        # Correct video aspect ratio for video rotation. i.e. aspect ratio
        # corresponds to the ratio after rotation.
        if abs(self._rotation) == 90 or abs(self._rotation == 270):
            video_ratio = 1/video_ratio

        # Tranform video to fill the video slideshow widget. Only videos with
        # the same orientation will be resized to fill the widget. Videos with a
        # different orientation will be resized to fit the widget.
        if self._resize == "fill":

            # Determine maximum dimension for widget with landscape orientation.
            if widget_ratio > 1:
                # Determine required maximum dimension for the rotation
                # transformation based on aspect ratios.
                if widget_ratio >= video_ratio and video_ratio >= 1:
                    max_dim = self.width
                elif widget_ratio >= video_ratio and video_ratio < 1:
                    max_dim = self.height
                elif widget_ratio < video_ratio and video_ratio >= 1:
                    max_dim = round(self.height*video_ratio)
                else:  # widget_ratio > video_ratio and video_ratio < 1
                    max_dim = round(self.width/video_ratio)
            # Determine maximum dimension for widget with portrait orientation.
            else:  # widget_ratio <= 1:
                if widget_ratio > video_ratio and video_ratio > 1:
                    max_dim = round(self.height*video_ratio)
                    # max_dim = self.width
                elif widget_ratio <= video_ratio and video_ratio >= 1:
                    max_dim = self.width
                    # max_dim = round(self.height*video_ratio)
                elif widget_ratio >= video_ratio and video_ratio <= 1:
                    # max_dim = self.height
                    max_dim = round(self.width/video_ratio)
                else:  # widget_ratio < video_ratio and video_ratio < 1
                    # max_dim = round(self.width/video_ratio)
                    max_dim = self.height

            # Set size of video widget to square with maximum dimension
            self._video.size = (max_dim, max_dim)
            # Adjust position of video widget within slideshow video widget
            # to center rotated video.
            self._video.x = round(self.x + (self.width - max_dim)/2)
            self._video.y = round(self.y + (self.height - max_dim)/2)

            # Apply rotation if not zero
            if self._rotation != 0:
                with self._video.canvas.before:
                    PushMatrix()
                    Rotate(angle=self._rotation, origin=self._video.center, axis=(0, 0, 1))
                with self._video.canvas.after:
                    PopMatrix()

        # Default is to fit the video to the canvas
        else:  # self._resize == "fit"

            # Rotate canvas if required.
            if (self._rotation != 0):
                # Determine required maximum dimension for the rotation
                # transformation based on aspect ratios.
                if widget_ratio > video_ratio and video_ratio > 1:
                    max_dim = round(self.height*video_ratio)
                elif widget_ratio <= video_ratio and video_ratio >= 1:
                    max_dim = self.width
                elif widget_ratio >= video_ratio and video_ratio <= 1:
                    max_dim = self.height
                else:  # widget_ratio < video_ratio and video_ratio < 1
                    max_dim = round(self.width/video_ratio)

                # Set size video widget to square with maximum dimension
                self._video.size = (max_dim, max_dim)
                # Adjust position of video widget within slideshow video widget
                # to center rotated video.
                self._video.x = round(self.x + (self.width - max_dim)/2)
                self._video.y = round(self.y + (self.height - max_dim)/2)

                # Apply rotation.
                with self._video.canvas.before:
                    PushMatrix()
                    Rotate(angle=self._rotation, origin=self._video.center, axis=(0, 0, 1))
                with self._video.canvas.after:
                    PopMatrix()
            else:
                self._video.size = self.size

        # Log debug information
#       Logger.debug(f"Video uuid: {self._file.uuid}")
#       Logger.debug(f"Video type: {self._file.type}")
#       Logger.debug(f"Video source: {self._file.source}")
#       Logger.debug(f"Video orientation: {self._file.orientation}")
#       Logger.debug(f"Video rotation: {self._file.rotation}")
#       Logger.debug(f"Total rotation: {self._rotation}")
#       Logger.debug(f"Widget width: {self.width}")
#       Logger.debug(f"Widget height: {self.height}")
#       Logger.debug(f"Widget aspect ratio: {widget_ratio}")
#       Logger.debug(f"max_dim: {max_dim}")
#       Logger.debug(f"Video width: {self._video.width}")
#       Logger.debug(f"Video height: {self._video.height}")
#       Logger.debug(f"Video aspect ratio: {video_ratio}")
#       Logger.debug(f"Video x: {self._video.x}")
#       Logger.debug(f"Video y: {self._video.y}")
#       Logger.debug(f"Video center: {self._video.center}")



"""Module for WebDAV repositories."""

import logging
import repository
import tempfile

from repository import ConfigError, IoError, check_param, check_valid_required
from webdav3.client import Client

from .file import RepositoryFile


class Repository(repository.Repository):
    """Repository with WebDAV file base."""

    # Required and valid configuration parameters
    CONF_REQ_KEYS = {'url', 'user', 'password', 'cache'}
    CONF_VALID_KEYS = {'root'} | CONF_REQ_KEYS

    def __init__(self, uuid, config, index=None):
        """Initialize the repository.

        :param uuid: UUID of repository
        :type name: str
        :param index: Optional file metadata index. Default is None.
        :type index: repository.Index
        :param config: Dictionary with repository Configuration
        :type config: dict
        :raises: repository.UuidError, repository.ConfigError
        """
        # Call constructor of parent class.
        repository.Repository.__init__(self, uuid, config, index)

        # Basic initialization.
        self._url = config['url']
        self._user = config['user']
        self._password = config['password']
        self._root = config.get('root', "/")

        # Create temporary directory for file caching.
        self._cache_dir = tempfile.TemporaryDirectory(dir=config['cache'], prefix=f"{uuid}-")

        # Open WebDav client session.
        options = {
         'webdav_hostname': self._url,
         'webdav_login':    self._user,
         'webdav_password': self._password
        }
        self._client = Client(options)

    def _check_config(self, config):
        """Check the repository configuration.

        :param config:
        :type config: dict
        :raises: repository.ConfigError
        """
        # Make sure all valid and required parameters have been specified.
        check_valid_required(config, self.CONF_VALID_KEYS, self.CONF_REQ_KEYS)
        # Check parameter values.
        check_param('url', config, is_str=True)
        check_param('user', config, is_str=True)
        check_param('password', config, is_str=True)
        check_param('cache', config, is_str=True)

    @property
    def cache_dir(self):
        """Return path of cache directory for the temporary storage of files.

        :return: path of cache directory
        :rtype: str
        """
        return self._cache_dir.name

    @property
    def client(self):
        """Return WebDAV client session of the repository.

        :return: WebDAV client session
        :rtype: webdav3.client.Client
        """
        return self._client

    @property
    def root(self):
        """Return WebDAV root directory of the repository.

        :return: WebDav root directory
        :rtype: str
        """
        return self._root

    def iterator(self, index_lookup=True, extract_metadata=True):
        """Provide iterator to traverse through files in the repository.

        :param index_lookup: Flag indicating whether file metadata shall be
            looked up from index. Default is True.
        :type index_lookup: bool
        :param extract_metadata: Flag indicating whether file metadata shall be
            extracted from file if not available from index. Default is True.
        :type extract_metadata: bool
        :return: file iterator
        :return type: repository.FileIterator
        """
        return FileIterator(self, index_lookup, extract_metadata)

    def file_by_uuid(self, uuid, index_lookup=True, extract_metadata=True):
        """Return file within the repository by its UUID.

        :param uuid: UUID of file
        :type uuid: str
        :param index_lookup: Flag indicating whether file metadata shall be
            looked up from index. Default is True.
        :type index_lookup: bool
        :param extract_metadata: Flag indicating whether file metadata shall be
            extracted from file if not available from index. Default is True.
        :type extract_metadata: bool
        :return: file with matching UUID
        :rtype: repository.RepositoryFile
        :raises: repository.UuidError, repository.IoError
        """
        return RepositoryFile(uuid, self, self._index, index_lookup, extract_metadata)


class FileIterator(repository.FileIterator):
    """Iterator which can be used to traverse through files in a webdav repository."""

    def __init__(self, rep, index_lookup=True, extract_metadata=True):
        """Initialize file iterator.

        :param rep: WebDAV repository
        :type rep: repository.webdav.repository
        :param index_lookup: Flag indicating whether file metadata shall be
            looked up from index. Default is True.
        :type index_lookup: bool
        :param extract_metadata: Flag indicating whether file metadata shall be
            extracted from file if not available from index. Default is True.
        :type extract_metadata: bool
        :raises: repository.IoError
        """
        # Basic initialization.
        self._rep = rep
        self._index_lookup = index_lookup
        self._extract_metadata = extract_metadata
        self._dir_list = []

        # Create iterator for list of root directory.
        try:
            self._file_list = rep.client.list(rep.root, get_info=True)
        except Exception as e:
            raise IoError(f"An exception occurred while listing the root directory. {e}", e)
        self._iterator = iter(self._file_list)

    def __next__(self):
        """Provide next file in iteration.

        :returns: next file
        :rtype: repository.webdav.RepositoryFile
        :raises: StopIteration
        """
        try:
            # Retrieve the next directory entry.
            entry = self._iterator.__next__()
            # Continue to retrieve entries if not a file.
            while entry['isdir']:
                # Save all sub-directories for later.
                self._dir_list.append(entry['path'])
                entry = self._iterator.__next__()

            # Construct relative path to root directory of the repository.
            uuid = entry['path']
            # Return the next file.
            logging.debug(f"Creating webdav repository file '{uuid}'.")
            return self._rep.file_by_uuid(uuid, self._index_lookup, self._extract_metadata)

        except StopIteration:
            if len(self._dir_list) > 0:
                # Start all over with last subdirectory in the list
                try:
                    dir = self._dir_list.pop()
                    self._file_list = self._rep.client.list(dir, get_info=True)
                except Exception as e:
                    raise IoError(f"An exception occurred while listing directiory '{dir}'. {e}", e)
                self._iterator = iter(self._file_list)
                return self.__next__()
            else:
                # Raise exception to indicate end of iteration otherwise
                raise StopIteration



"""Package providing rclone file repositories.

Author: Bernd Kalbfuss
License: GNU General Public License v3 (GPLv3)
"""

from . file import RepositoryFile
from . repository import Repository, FileIterator
